
package knapsack;

import java.util.Random;


/**
 * Evolutionary knapsack load planner.
 *
 * @author	Copyright (C) 2002 Alessandro Bollini
 */
public class Planner {

	// Input data.

	private final int weightLimit=35; //[kg]
	private final int volumeLimit=80; // [lt]

	private final Inventory inventory=new Inventory();

	// Process configuration.

	private final long seed=System.currentTimeMillis();
	private final Random randomizer=new Random(seed);

	private final int populationSize=100000;
	private final int tournamentSize=2;

	private final double recombinationRate=0.7;
	private final double mutationRate=0.1;

	private final int iterationLimit=50;
	private final int targetValue=2000;

	// Process state.

	private Plan[] population;

	private int iteration;
	private Plan solution;


	public static void main(final String[] args) {

		if ( args.length > 0 ) {
			throw new IllegalArgumentException("unexpected arguments");
		}

		new Planner().run();
	}


	private Planner() {

		population=new Plan[populationSize];

		iteration=0;
		solution=new Plan(
			new boolean[inventory.getSize()],
			new Knapsack(weightLimit, volumeLimit), 0.0);

		for ( int n=0; n < population.length; ++n ) {

			final boolean[] genotype=generate();
			final Plan individual=develop(genotype);

			population[n]=individual;

			if ( individual.getFitness() > solution.getFitness() ) {
				solution=individual;
			}
		}
	}

	private boolean[] generate() {

		final boolean[] genotype=new boolean[inventory.getSize()];
		genotype[randomizer.nextInt(genotype.length)]=true;

		return genotype;
	}

	private Plan develop(final boolean[] genotype) {

		final Knapsack phenotype=express(genotype);
		final double fitness=evaluate(phenotype);

		return new Plan(genotype, phenotype, fitness);
	}

	private Knapsack express(final boolean[] genotype) {

		final Knapsack knapsack=new Knapsack(weightLimit, volumeLimit);

		try {

			for ( int n=0; n < genotype.length; ++n ) {
				if ( genotype[n] == true ) {
					knapsack.add(inventory.getItem(n));
				}
			}

		} catch ( final IllegalStateException exception ) {
			knapsack.empty();
		}

		return knapsack;
	}

	private double evaluate(final Knapsack phenotype) {
		return phenotype.getValue();
	}


	private void run() {

		System.out.println("------------------------------ Configuration --");
		System.out.println("seed\t"+seed);

		System.out.println("-------------------------------------- Trace --");
		System.out.println("iter.\tweight\tvolume\tvalue");

		while ( !completed() ) {

			System.out.println(
				iteration
				+"\t\t"+solution.getPhenotype().getWeight()
				+"\t\t"+solution.getPhenotype().getVolume()
				+"\t\t"+solution.getPhenotype().getValue()
			);

			iterate();
		}

		System.out.println("----------------------------------- Knapsack --");
		System.out.println(solution.getPhenotype());
	}


	private boolean completed() {
		return solution.getFitness() >= targetValue
			|| iteration >= iterationLimit;
	}

	private void iterate() {

		for (int n=0; n < population.length; ++n ) {

			final Plan parent=select();
			final Plan offspring=reproduce(parent);
			final int target=randomizer.nextInt(population.length);

			population[target]=offspring;

			if ( offspring.getFitness() > solution.getFitness() ) {
				solution=offspring;
			}
		}

		++iteration;
	}


	private Plan select() {

		Plan winner=sample();

		for ( int n=0; n < tournamentSize; ++n ) {

			final Plan challenger=sample();

			if ( challenger.getFitness() > winner.getFitness() ) {
				winner=challenger;
			}
		}

		return winner;
	}

	private Plan sample() {
		return population[randomizer.nextInt(population.length)];
	}


	private Plan reproduce(final Plan individual) {

		final double event=randomizer.nextDouble();
		final boolean[] genotype;

		if ( event < recombinationRate ) {

			final Plan peer=select();

			genotype=recombine(
				individual.getGenotype(), peer.getGenotype());

		} else if ( event-recombinationRate < mutationRate ) {

			genotype=mutate(individual.getGenotype());

		} else {

			genotype=replicate(individual.getGenotype());
		}

		return develop(genotype);
	}

	private boolean[] recombine(
		final boolean[] targetGenotype, boolean[] sourceGenotype) {

		final boolean[] combinedGenotype=(boolean[])targetGenotype.clone();
		final int locus=randomizer.nextInt(combinedGenotype.length-1)+1;

		System.arraycopy(sourceGenotype, locus,
			combinedGenotype, locus,
			combinedGenotype.length-locus);

		return combinedGenotype;

	}

	private boolean[] mutate(final boolean[] targetGenotype) {

		final boolean[] mutatedGenotype=(boolean[])targetGenotype.clone();
		final int locus=randomizer.nextInt(mutatedGenotype.length);

		mutatedGenotype[locus]=!mutatedGenotype[locus];

		return mutatedGenotype;
	}

	private boolean[] replicate(final boolean[] targetGenotype) {
		return (boolean[])targetGenotype.clone();
	}
}
